//=============================================================================
// ProMesh.NET - .NET Web Application Framework 
//
// Copyright (c) 2003-2008 Philippe Leybaert
//
// Permission is hereby granted, free of charge, to any person obtaining a copy 
// of this software and associated documentation files (the "Software"), to deal 
// in the Software without restriction, including without limitation the rights 
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
// copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in 
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//=============================================================================

using System;
using System.Collections.Generic;
using System.Configuration;
using System.Reflection;
using System.Text.RegularExpressions;

namespace Activa.ProMesh
{
    public static class WebAppConfig
    {
        private static Type                          _sessionType = typeof(SessionBase);

        private static ITranslationProvider          _translationProvider;
        private static ISessionLoggingProvider       _sessionLoggingProvider = new MinimalSessionLoggingProvider();
        private static IVisitorProvider              _visitorProvider = new MinimalVisitorProvider();
        private static IPageLoggingProvider          _loggingProvider;
        private static IAjaxProvider                 _ajaxProvider = new JQueryAjaxProvider();
        private static ITemplateResolver             _templateResolver;
        private static ISessionSerializer            _sessionSerializer;

        private static Dictionary<string, ControllerClass> _controllerTypesByName;
        private static Dictionary<string, ControllerClass> _viewComponentsByName;
        private static Dictionary<Type, ControllerClass> _controllerTypesByType;

        private static readonly List<Assembly>                 _registeredAssemblies = new List<Assembly>();
        private static readonly List<ICustomObjectCreator>     _customObjectCreators = new List<ICustomObjectCreator>();

        private static string _templatePath          = "templates";
        private static string _defaultLayout         = "master";
        private static string _defaultLanguage       = "en";
        private static string _pageExtension         = ".ashx";
        private static bool   _useLanguagePath;
        private static bool   _forceLanguagePath;

        private static readonly string _version;

        public static event Action<Exception> ExceptionOccurred;
        public static event Action<string> PageNotFound;
        public static event Action<SessionBase> SessionCreated;

        private static readonly Router _router = new Router();

        static WebAppConfig()
        {
            Regex.CacheSize = 100;

            _version = Assembly.GetExecutingAssembly().GetName().Version.ToString();

            string applicationClassName = null;

            foreach (string configKey in ConfigurationManager.AppSettings.Keys)
            {
                string configValue = ConfigurationManager.AppSettings[configKey];

                switch (configKey.ToLower())
                {
                    case "promesh.defaultlayout":
                        DefaultLayout = configValue;
                        break;
                    case "promesh.templatepath":
                        TemplatePath = configValue;
                        break;
                    case "promesh.pageextension":
                        PageExtension = configValue;
                        break;
                    case "promesh.uselanguagepath":
                        UseLanguagePath = (configValue.ToLower() == "true");
                        break;
                    case "promesh.forcelanguagepath":
                        ForceLanguagePath = (configValue.ToLower() == "true");
                        break;
                    case "promesh.defaultlanguage":
                        DefaultLanguage = configValue.ToUpper();
                        break;
                    case "promesh.applicationclass":
                        applicationClassName = configValue;
                        break;
                    case "promesh.usedefaultroutes":
                        _router.UseDefaultRoutes = (configValue.ToLower() == "true");
                        break;
                }
            }

            if (applicationClassName == null)
                throw new Exception("No ProMesh.ApplicationClass defined in web.config");

            Type appType = Type.GetType(applicationClassName, false);

            if (appType == null)
                throw new Exception("Application class {" + applicationClassName + "} could not be loaded");

            MethodInfo initMethod = appType.GetMethod("Init", new Type[0]);

            if (initMethod == null || !initMethod.IsStatic)
                throw new Exception("No \"public static void Init()\" method defined for class " + appType.Name);

            RegisterAssembly(appType.Assembly);
            RegisterAssembly(Assembly.GetExecutingAssembly());

            initMethod.Invoke(null, null);

            LoadControllerClasses();

            _router.AddDefaultRoutes();
        }

        public static void Init()
        {
            // This method does nothing, but it will trigger the execution of the static constructor for WebApp
        }

        public static Type SessionType
        {
            get
            {
                return _sessionType;
            }
            set
            {
                if (!typeof(SessionBase).IsAssignableFrom(value))
                    throw new Exception("SessionType should be derived from SessionBase");

                _sessionType = value;
            }
        }

        public static ITranslationProvider TranslationProvider
        {
            get { return _translationProvider; }
            set { _translationProvider = value; }
        }

        public static ISessionLoggingProvider SessionLoggingProvider
        {
            get { return _sessionLoggingProvider; }
            set { _sessionLoggingProvider = value; }
        }

        public static IPageLoggingProvider LoggingProvider
        {
            get { return _loggingProvider; }
            set { _loggingProvider = value; }
        }

        public static IAjaxProvider AjaxProvider
        {
            get { return _ajaxProvider; }
            set { _ajaxProvider = value; }
        }

        public static ITemplateResolver TemplateResolver
        {
            get { return _templateResolver; }
            set { _templateResolver = value; }
        }

        public static ISessionSerializer SessionSerializer
        {
            get { return _sessionSerializer; }
            set { _sessionSerializer = value; }
        }

        internal static List<ICustomObjectCreator> CustomObjectCreators
        {
            get { return _customObjectCreators; }
        }

        public static string TemplatePath
        {
            get { return _templatePath; }
            set { _templatePath = value; }
        }

        public static string DefaultLayout
        {
            get { return _defaultLayout; }
            set { _defaultLayout = value; }
        }

        public static string DefaultLanguage
        {
            get { return _defaultLanguage; }
            set { _defaultLanguage = value; }
        }

        public static string PageExtension
        {
            get { return _pageExtension; }
            set 
            { 
                if (string.IsNullOrEmpty(value)) 
                    _pageExtension = ""; 
                else if (!value.StartsWith(".")) 
                    _pageExtension = "." + value; 
                else
                    _pageExtension = value;
            }
        }

        public static bool UseLanguagePath
        {
            get { return _useLanguagePath; }
            set { _useLanguagePath = value; }
        }

        public static bool ForceLanguagePath
        {
            get { return _forceLanguagePath; }
            set { _forceLanguagePath = value; }
        }

        public static string Version
        {
            get { return _version; }
        }

        public static IVisitorProvider VisitorProvider
        {
            get { return _visitorProvider; }
            set { _visitorProvider = value; }
        }

        public static Router Router
        {
            get { return _router; }
        }

        public static void RegisterAssembly(Assembly assembly)
        {
            // We don't need to load anything, because if the assembly object is here, the assembly is automatically loaded

            if (_registeredAssemblies.Find(delegate(Assembly a) { return a.FullName == assembly.FullName; }) == null)
            {
                _registeredAssemblies.Add(assembly);
            }
        }

        public static void RegisterAssembly(string assemblyPath)
        {
            RegisterAssembly(Assembly.LoadFrom(assemblyPath));
        }

        private static void LoadControllerClasses()
        {
            _controllerTypesByName = new Dictionary<string, ControllerClass>(StringComparer.OrdinalIgnoreCase);
            _viewComponentsByName = new Dictionary<string, ControllerClass>(StringComparer.OrdinalIgnoreCase);
            _controllerTypesByType = new Dictionary<Type, ControllerClass>();

            List<Type> controllerTypes = new List<Type>();

            foreach (Assembly assembly in _registeredAssemblies)
            {
                controllerTypes.AddRange(TypeHelper.FindCompatibleTypes(assembly, typeof (Controller)));
            }

            foreach (Type type in controllerTypes)
            {
                ControllerClass controllerClass = new ControllerClass(type);

                if (controllerClass.IsViewComponent)
                {
                    _viewComponentsByName.Add(controllerClass.Name, controllerClass);
                }
                else
                {
                    _controllerTypesByName.Add(controllerClass.Name, controllerClass);
                    _controllerTypesByType.Add(type, controllerClass);

                    foreach (Route route in controllerClass.Routes)
                        Router.AddRoute(route);
                }
            }
        }

        internal static ControllerClass GetControllerClass(string controllerName)
        {
            ControllerClass controllerClass;

            if (_controllerTypesByName.TryGetValue(controllerName, out controllerClass))
                return controllerClass;

            return null;
        }

        internal static ControllerClass GetControllerClass(Type controllerType)
        {
            ControllerClass controllerClass;

            if (_controllerTypesByType.TryGetValue(controllerType, out controllerClass))
                return controllerClass;

            return null;
        }

        internal static ControllerClass GetViewComponentClass(string viewComponentName)
        {
            ControllerClass controllerClass;

            if (_viewComponentsByName.TryGetValue(viewComponentName, out controllerClass))
                return controllerClass;

            return null;
        }

        public static void RegisterCustomObjectCreator(ICustomObjectCreator creator)
        {
            _customObjectCreators.Add(creator);
        }

        internal static bool Fire_ExceptionHandler(Exception ex)
        {
            if (ExceptionOccurred != null)
            {
                ExceptionOccurred(ex);

                return true;
            }

            return false;
        }

        internal static bool Fire_PageNotFound(string pageUrl)
        {
            if (PageNotFound != null)
            {
                PageNotFound(pageUrl);

                return true;
            }

            return false;
        }

        internal static bool Fire_SessionCreated(SessionBase session)
        {
            session.OnSessionCreated();

            if (SessionCreated != null)
            {
                SessionCreated(session);

                return true;
            }

            return false;
        }
    }
}
